#include "blitz_array.h"
#include "blitz_string.h"

typedef uint_fast32_t	strpos_t;
typedef uint_fast32_t	arrpos_t;

static void array_free(BBArray *arr);
static bool array_sub_equals(const BBArray * restrict arr, const BBArray * restrict rhs,
							 BBSize start_pos, BBSize len)
							__attribute__ ((pure, always_inline, nonnull(1, 2), optimize(3)));

static BBDebugScope ClassScope =
{
	.kind			= BBDEBUGSCOPE_USERTYPE,
	.name			= "Array",
	.decls			= {
		BBDEBUGDECL_END
	}
};

BBClass bbArrayClass =
{
	.super			= &bbObjectClass,
	.free			= array_free,
	.debug_scope	= &ClassScope,
	.instance_size	= 0,
	
	.ctor			= 0,
	.dtor			= 0,
	
	.ToString		= bbArrayToString,
	.Compare			= bbArrayCompare,
	.SendMessage		= bbObjectSendMessage,
	.Clone			= bbArrayClone,
	
	._reserved1_		= bbObjectReserved,
	._reserved2_		= bbObjectReserved,
	
	.vfns			= {
		bbArraySort,
		bbArrayDimensions
	}
};

BBArray bbEmptyArray =
{
	.clas		= &bbArrayClass,
	.refs		= BBGC_MANYREFS,
	
	.type		= "",
	.dims		= 0,
	.size		= 0,
	.scales[0]	= 0
};

// Only used by ref counting GC.
static void array_free(BBArray *arr)
{
#ifdef BB_GC_RC
	if (arr == &bbEmptyArray) {
		arr->refs = BBGC_MANYREFS;
		return;
	}
	
	// Complex types must be dereferenced in ref-counting GC
	if (BBARRAY_ISCOMPLEX(arr)) {
		
		BBObject **data = (BBObject **)BBARRAY_DATAPTR(arr);
		BBObject *obj;
		
		for (BBSize i = arr->scales[0]; i > 0; --i) {
			obj = *data++;
			BB_DECREFS(obj);
		}
	}
	
	bbGCDeallocObject(arr, BBARRAY_SIZE(arr));
#endif
}

// Contains the GC allocation flags indexed by BBTypeKey
static const BBInt ArrayAllocFlags[] =
{
	[BBTYPEKEY_VOID]		= BBGC_ATOMIC,
	[BBTYPEKEY_BYTE]		= BBGC_ATOMIC,
	[BBTYPEKEY_SHORT]	= BBGC_ATOMIC,
	[BBTYPEKEY_INT]		= BBGC_ATOMIC,
	[BBTYPEKEY_LONG]		= BBGC_ATOMIC,
	[BBTYPEKEY_FLOAT]	= BBGC_ATOMIC,
	[BBTYPEKEY_DOUBLE]	= BBGC_ATOMIC,
	
	[BBTYPEKEY_POINTER]	= BBGC_ATOMIC,
	[BBTYPEKEY_FUNCTION]	= BBGC_ATOMIC,	
	
	[BBTYPEKEY_OBJECT]	= 0,
	[BBTYPEKEY_STRING]	= 0,
	[BBTYPEKEY_ARRAY]	= 0
};


BBArray* bb_array_alloc(type, dims, lens)
 const char *type;
 const BBSize dims;
 const BBSize *lens;
{
	BBSize length = 1;
	BBInt i;
	
	for (i = 0; i < dims; ++i)
	{
		if (lens[i] < 1) {
			return &bbEmptyArray;
		}
		
		length *= lens[i];
	}
	
	const BBSize size = BBTYPETAG_SIZE(type) * length;

	BBArray *arr = BBARRAY_ALLOC(size, dims, ArrayAllocFlags[BBTYPETAG_KEY(type)]);
	
	arr->type = type;
	arr->dims = dims;
	arr->size = size;

	for (i = 0; i < dims; ++i)
	{
		arr->scales[i] = lens[i];
	}
	
	for (i = dims - 2; i >= 0; --i)
	{
		arr->scales[i] *= arr->scales[i + 1];
	}

	return arr;
}

BBArray* bb_array_alloc1d(type, len)
 const char *type;
 const BBSize len;
{
	const BBSize size = len * BBTYPETAG_SIZE(type);
	BBArray *arr = BBARRAY_ALLOC(size, 1, ArrayAllocFlags[BBTYPETAG_KEY(type)]);
	
	arr->type = type;
	arr->dims = 1;
	arr->size = size;
	arr->scales[0] = len;
	
	return arr;
}

void bb_array_init(arr)
 BBArray *arr;
{
	if (arr->scales[0] < 1) {
		return;
	}
	
	void **arr_data  = BBARRAY_DATAPTR(arr);
	const void *init = BBTYPETAG_INIT(arr->type);
	
	if (init) {
		
		for (BBSize i = arr->scales[0]; i > 0; --i)
		{
			*arr_data++ = init;
		}
		
	} else {
		memset(arr_data, 0, arr->size);
	}
}

static inline bool array_sub_equals(arr, rhs, start_pos, cmp_len)
 const BBArray * restrict arr;
 const BBArray * restrict rhs;
 BBSize start_pos;
 const BBSize cmp_len;
{
	const BBSize arr_len = arr->scales[0];
	const BBSize rhs_len = rhs->scales[0];
	
	if (arr_len < rhs_len ||
		start_pos + cmp_len > arr_len ||
		cmp_len > rhs_len) {
		return false;
	}
	
	const BBTypeKey arr_tkey = BBTYPETAG_KEY(arr->type);
	const BBTypeKey rhs_tkey = BBTYPETAG_KEY(rhs->type);
	
	arrpos_t pos;
	
	// Only compare arrays of equal type.
	if (arr_tkey == rhs_tkey) {
		if (BBTYPEKEY_ISPRIMITIVE(arr_tkey)) {
			
			return !memcmp(BBARRAY_DATAPTR(arr), BBARRAY_DATAPTR(rhs),
						   rhs_len  * BBTYPEKEY_SIZE(arr_tkey));
			
		} else if (arr_tkey == BBTYPEKEY_STRING) {
			
			const BBString **arr_data = ((BBStringArray *)arr)->data;
			const BBString **rhs_data = ((BBStringArray *)rhs)->data;
			
			for (pos = 0; pos < cmp_len; ++pos)
			{
				if (!bbStringEquals(arr_data[start_pos + pos], rhs_data[pos])) {
					return false;
				}
			}
			
		} else if (arr_tkey == BBTYPEKEY_OBJECT) {
			
			const BBObject **arr_data = ((BBObjectArray *)arr)->data;
			const BBObject **rhs_data = ((BBObjectArray *)rhs)->data;
			
			arr_data += start_pos;
			
			BBClassCompareFunc cmp_func;
			
			for (pos = 0; pos < cmp_len; ++pos)
			{
				cmp_func = arr_data[pos]->clas->Compare;
				
				if (cmp_func(arr_data[pos], rhs_data[pos])) {
					return false;
				}
			}
			
		} else if (arr_tkey == BBTYPEKEY_ARRAY) {
			
			const BBArray **arr_data = ((BBArrayArray *)arr)->data;
			const BBArray **rhs_data = ((BBArrayArray *)rhs)->data;
			
			arr_data += start_pos;
			
			for (pos = 0; pos < cmp_len; ++pos)
			{
				if (!bbArrayEquals(arr_data[pos], rhs_data[pos])) {
					return false;
				}
			}
		}
	}
	
	return true;
}

//	API //

BBInt bbArrayCompare(const BBArray *arr, const BBArray *rhs)
{
	if (arr == rhs) {
		return 0;
	}
	
	if (arr->scales[0] != rhs->scales[0]) {
		return arr->scales[0] > rhs->scales[0] ? 1 : -1;
	} else if (arr->dims != rhs->dims) {
		return arr->dims > rhs->dims ? 1 : -1;
	} else if (arr->size != rhs->size) {
		return arr->size > rhs->size ? 1 : -1;
	}
	
	const BBTypeKey arr_typekey = BBTYPETAG_KEY(arr->type);
	const BBTypeKey rhs_typekey = BBTYPETAG_KEY(rhs->type);
	
	if (arr_typekey != rhs_typekey) {
		return arr_typekey > rhs_typekey ? 1 : -1;
	}
	
	int cmp = memcmp(BBARRAY_DATAPTR(arr), BBARRAY_DATAPTR(rhs), arr->size);
	if (cmp) {
		return cmp > 0 ? 1 : -1;
	}
	
	return 0;
}

BBArray *bbArrayNew(const char *type, BBInt dims, ...)
{
#if BB_ARGP
	BBInt *lens = (BBInt *)bbArgp(8);
#else
	BBInt *lens = &dims + 1;
#endif

	BB_ASSERT(lens[0] >= 0, "Array length cannot be negative");
	BB_ASSERT(lens[0] < (BBARRAY_MAX_LENGTH / BBTYPETAG_SIZE(type)), "Array is too long");		
	
	BBArray *out = bb_array_alloc(type, dims, lens);
	bb_array_init(out);
	
	return out;
}
 
BBArray* bbArrayNewEx(const char *type, BBInt dims, BBInt *lens)
{
	BB_ASSERT(lens[0] >= 0, "Array length cannot be negative");
	BB_ASSERT(lens[0] < (BBARRAY_MAX_LENGTH / BBTYPETAG_SIZE(type)), "Array is too long");	
	
	BBArray *out = bb_array_alloc(type, dims, lens);
	bb_array_init(out);
	
	return out;
}

// Manually inlined, since this is used most often
BBArray* bbArrayNew1D(const char *type, BBInt len)
{
	const BBTypeKey tkey	= BBTYPETAG_KEY(type);
	const BBSize size		= len * BBTYPEKEY_SIZE(tkey);

	BB_ASSERT(len >= 0, "Array length cannot be negative");
	BB_ASSERT(len < (BBARRAY_MAX_LENGTH / BBTYPEKEY_SIZE(tkey)), "Array is too long");
	
	BBArray *arr	= BBARRAY_ALLOC(size, 1, ArrayAllocFlags[tkey]);
	
	arr->type		= type;
	arr->dims		= 1;
	arr->size		= size;
	arr->scales[0]	= len;
	
	void **arr_data  = BBARRAY_DATAPTR(arr);
	const void *init = BBTYPEKEY_INIT(tkey);
	
	if (init) {
		
		for (BBSize i = len; i > 0; --i)
		{
			*arr_data++ = init;
		}
		
	} else {
		memset(arr_data, 0, size);
	}
	
	return arr;
}

BBArray* bbArraySlice(const char *type, BBArray *in_arr, BBInt beg, BBInt end)
{
	const int length = end - beg;
	if (length < 1) {
		return &bbEmptyArray;
	}

	BBArray *out_arr = bb_array_alloc1d(type, length);

	const BBSize element_size = out_arr->size / length;
	
	const void * const initializer = BBTYPETAG_INIT(out_arr->type);
	char *data_ptr = (char *)BBARRAY_DATAPTR_CALC(out_arr, 1);

	int k;
	
	int n = -beg;
	if (n > 0) {
		
		if (beg + n > end) {
			n = end - beg;
		}
		
		if (initializer) {
			void **dst = (void **)data_ptr;
			
			for (k = 0; k < n; ++k)
			{
				*dst++ = initializer;
			}
			
			data_ptr = (char*)dst;
		} else {
			memset(data_ptr, 0, n * element_size);
			data_ptr += n * element_size;
		}
		
		beg += n;
		if (beg == end) {
			return out_arr;
		}
	}
	
	n = in_arr->scales[0] - beg;
	if (n > 0) {
		if (beg + n > end) {
			n = end - beg;
		}
#ifdef BB_GC_RC
		if (BBARRAY_TYPE_ISCOMPLEX(type)) {
			BBObject **dst = (BBObject **)data_ptr;
			BBObject **src = (BBObject **)BBARRAY_DATAPTR(in_arr) + beg;
			
			for (k = 0; k < n; ++k)
			{ 
				BBObject *o = *src++;
				BB_INCREFS(o);
				*dst++ = o; 
			}
			
			data_ptr = (char *)dst;
		} else {
			memcpy(
				data_ptr,
				(char *)BBARRAY_DATAPTR(in_arr) + beg * element_size,
				n * element_size
			);
			
			data_ptr += n * element_size;
		}
#else
		memcpy(
			data_ptr,
			(char *)BBARRAY_DATAPTR(in_arr) + beg * element_size,
			n * element_size
		);
		
		data_ptr += n * element_size;
#endif
		beg += n;
		if (beg == end) {
			return out_arr;
		}
	}
	
	n = end - beg;
	if (n > 0) {
		if (initializer) {
			void **dst = (void **)data_ptr;
			for (k = 0; k < n; ++k)
			{
				*dst++ = initializer;
			}
		} else {
			memset(data_ptr, 0, n * element_size);
		}
	}
	
	return out_arr;
}

BBArray* bbArrayConcat(const char *type, BBArray *x, BBArray *y)
{
	const BBSize out_len = x->scales[0] + y->scales[0];
	
	if (out_len < 1) {
		return &bbEmptyArray;
	}
	
	BBArray *out = bb_array_alloc1d(type, out_len);
	char *data = (char *)BBARRAY_DATAPTR_CALC(out, 1);
	
	memcpy(data, BBARRAY_DATAPTR_CALC(x, 1), x->size);
	memcpy(data + x->size, BBARRAY_DATAPTR_CALC(y, 1), y->size);
	
#ifdef BB_GC_RC
	// Increasing the reference count of all complex data type elements
	// is necessary to let the GC know that the new array is using them.
	if (BBARRAY_TYPE_ISCOMPLEX(type)) {
		
		BBObject **objects = (BBObject **)data;
		
		for (arrpos_t i = 0; i < out_len; ++i)
		{
			BBObject *obj = *objects++;
			BB_INCREFS(obj);
		}
	}
#endif	// BB_GC_RC
	
	return out;
}

BBArray* bbArrayFromData(const char *type, BBInt length, const void *data)
{
	if (length < 1) {
		return &bbEmptyArray;
	}
	
	BBArray *arr = bb_array_alloc1d(type, length);

	if (type[0] == 'b') {
		
		BBShort *arr_data = ((BBByteArray *)arr)->data;
		
		for (arrpos_t k = 0; k < length; ++k) {
			arr_data[k] = ((int *)data)[k];
		}
		
	} else if (type[0] == 's') {
		
		BBShort *arr_data = ((BBShortArray *)arr)->data;
		
		for (arrpos_t k = 0; k < length; ++k) {
			arr_data[k] = ((int *)data)[k];
		}
		
	} else {
		memcpy(BBARRAY_DATAPTR_CALC(arr, 1), data, arr->size);
	}
 
	return arr;
}

BBArray* bbArrayDimensions(const BBArray *arr)
{
	if (arr->scales[0] < 1) {
		return &bbEmptyArray;
	}
		
	BBSize num_dims = arr->dims;
	
	BBArray *out	= bb_array_alloc1d("i", num_dims);
	BBInt *out_data	= ((BBIntArray *)out)->data;
	
	--num_dims;
	
	for (arrpos_t i = 0; i < num_dims; ++i) {
		*++out_data = arr->scales[i] / arr->scales[i+1];
	}
	
	*++out_data = arr->scales[num_dims - 1];
	
	return out;
}

BBArray* bbArrayCastFromObject(const BBObject *obj, const char *typetag)
{
	const BBArray *arr = (BBArray *)obj;
	
	if (arr == &bbEmptyArray) {
		return arr;
	}
	
	if (arr->clas != &bbArrayClass) {
		return (BBArray *)BBNULLOBJECT;
	}
	
	if (arr->type[0] == ':' && typetag[0] == ':') {
		return arr;
	}
	
	if (strcmp(arr->type, typetag)) {
		return (BBArray *)BBNULLOBJECT;
	}
	
	return arr;
}

#define SWAP(x, y)	{ t = *(x); *(x) = *(y); *(y) = t; }
#define QSORTARRAY( TYPE,IDENT )\
static void IDENT( TYPE *lo,TYPE *hi ){\
	TYPE t;\
	TYPE *i;\
	TYPE *x;\
	TYPE *y;\
	if( hi<=lo ) return;\
	if( lo+1==hi ){\
		if( LESSTHAN(hi,lo) ) SWAP(lo,hi);\
		return;\
	}\
	i=(hi-lo)/2+lo;\
	if( LESSTHAN(i,lo) ) SWAP(i,lo);\
	if( LESSTHAN(hi,i) ){\
		SWAP(i,hi);\
		if( LESSTHAN(i,lo) ) SWAP(i,lo);\
	}\
	x=lo+1;\
	y=hi-1;\
	do{\
		while( LESSTHAN(x,i) ) ++x;\
		while( LESSTHAN(i,y) ) --y;\
		if( x>y ) break;\
		if( x<y ){\
			SWAP(x,y);\
			if( i==x ) i=y;\
			else if( i==y ) i=x;\
		}\
		++x;\
		--y;\
	}while( x<=y );\
	IDENT(lo,y);\
	IDENT(x,hi);\
}

#undef LESSTHAN
#define LESSTHAN(X,Y) (*(X) < *(Y))
QSORTARRAY(BBByte, _qsort_b)
QSORTARRAY(BBShort, qsort_s)
QSORTARRAY(BBInt, qsort_i)
QSORTARRAY(BBLong, qsort_l);
QSORTARRAY(BBFloat, qsort_f);
QSORTARRAY(BBDouble, qsort_d);
#undef LESSTHAN
#define LESSTHAN(X,Y) ((*X)->clas->Compare(*(X),*(Y)) < 0)
QSORTARRAY( BBObject*,qsort_obj );
#undef LESSTHAN
#define LESSTHAN(X,Y) (*(X)>*(Y))
QSORTARRAY(BBByte, qsort_b_d)
QSORTARRAY(BBShort, qsort_s_d)
QSORTARRAY(BBInt, qsort_i_d)
QSORTARRAY(BBLong, qsort_l_d);
QSORTARRAY(BBFloat, qsort_f_d);
QSORTARRAY(BBDouble, qsort_d_d);
#undef LESSTHAN
#define LESSTHAN(X,Y) ((*X)->clas->Compare(*(X),*(Y)) > 0)
QSORTARRAY( BBObject*,qsort_obj_d );

void bbArraySort(BBArray *arr, BBInt ascending)
{
	int n = arr->scales[0] - 1;
	if (n <= 0) {
		return;
	}
	
	void *p = BBARRAY_DATAPTR(arr);
	
	if (ascending) {
		
		switch (arr->type[0])
		{
		case 'b':
			_qsort_b((BBByte *)p, (BBByte *)p + n);
			break;
		case 's':
			qsort_s((BBShort *)p, (BBShort *)p + n);
			break;
		case 'i':
			qsort_i((BBInt *)p, (BBInt *)p + n);
			break;
		case 'l':
			qsort_l((BBLong *)p, (BBLong *)p + n);
			break;
		case 'f':
			qsort_f((BBFloat *)p, (BBFloat *)p + n );
			break;
		case 'd':
			qsort_d((BBDouble *)p, (BBDouble *)p + n);
			break;
		case '$': case ':':
			qsort_obj((BBObject **)p, (BBObject **)p + n);
			break;
		}
		
	} else {
		
		switch (arr->type[0]) {
		case 'b':
			qsort_b_d((BBByte *)p, (BBByte *)p + n);
			break;
		case 's':
			qsort_s_d((BBShort *)p, (BBShort *)p + n);
			break;
		case 'i':
			qsort_i_d((BBInt *)p, (BBInt *)p + n);
			break;
		case 'l':
			qsort_l_d((BBLong *)p,(BBLong *)p + n);
			break;
		case 'f':
			qsort_f_d((BBFloat *)p,(BBFloat *)p + n);
			break;
		case 'd':
			qsort_d_d((BBDouble *)p, (BBDouble *)p + n);
			break;
		case '$': case ':':
			qsort_obj_d((BBObject **)p, (BBObject **)p + n);
			break;
		}
	}
}

/// EXTENDED ///

BBArray* bbArrayFromCArray(buf, size, tkey, dims)
 const void *buf;
 const BBInt size;
 const BBTypeKey tkey;
 const BBIntArray *dims;
{
	if (!buf || size < 1 || dims->dims > 1 || !BBTYPEKEY_ISNUMERIC(tkey)) {
		return &bbEmptyArray;
	}
	
	const int num_elements = size / BBTYPEKEY_SIZE(tkey);
	
	BBArray *out;
	
	if (dims->scales[0]) {
		out = bb_array_alloc(BBTYPEKEY_TAG(tkey), dims->scales[0], dims->data);
	} else {
		out = bb_array_alloc1d(BBTYPEKEY_TAG(tkey), num_elements);
	}
	
	memcpy(BBARRAY_DATAPTR(out), buf, out->size);
			
	return out;
}

BBIntArray* bbArrayFromInts(const BBInt *buf, const BBInt size)
{
	if (!buf || size < 1) {
		return &bbEmptyArray;
	}
	
	BBIntArray* out = bb_array_alloc1d("i", size);
	memcpy(out->data, buf, size * sizeof (BBInt));
	
	return out;
}

BBFloatArray* bbArrayFromFloats(const BBFloat *buf, const BBInt size)
{
	if (!buf || size < 1) {
		return &bbEmptyArray;
	}
	
	BBIntArray* out = bb_array_alloc1d("f", size);
	memcpy(out->data, buf, size * sizeof (BBFloat));
	
	return out;
}


static const BBChar EmptyArrayMessageBuf[] = {
	'(','A','r','r','a','y',')',
	' ','N','u','l','l'
};

#define TOSTRING_NUMBUF_SIZE	32

// Still a bit buggy .. apparently the array properties are not
// changed when its resized or shit .. I'm not awake :D
BBString* bbArrayToString(const BBArray *arr)
{
	if (!arr || arr == BBNULLARRAY) {
		return BBSTRING_FROMBUFFER(EmptyArrayMessageBuf, 12);
	}
	
	const BBTypeKey tkey = BBTYPETAG_KEY(arr->type);

	char *ident			= bb_typetag_identifier(arr->type);
	size_t ident_len		= strlen(ident);
				
	size_t buf_len	= ident_len + 256 + arr->dims * TOSTRING_NUMBUF_SIZE;
	char *buf 		= malloc(buf_len);
	
	snprintf(buf, buf_len, "(Array) %s[", ident);
	free(ident);
	
	char num_buf[TOSTRING_NUMBUF_SIZE];
	const BBSize num_dims = arr->dims;
	
	if (num_dims == 1) {
		snprintf(buf, buf_len, "%s%i]", buf, arr->scales[0]);
	} else {
		
		for (BBSize d = 0; d < num_dims - 1; ++d)
		{
			snprintf(num_buf, TOSTRING_NUMBUF_SIZE, "%i,", arr->scales[d] / arr->scales[d+1]);
			strncat(buf, num_buf, buf_len);
		}
		
		snprintf(num_buf, TOSTRING_NUMBUF_SIZE, "%i]", arr->scales[arr->dims-1]);
		strncat(buf, num_buf, buf_len);
	}	
	
	const size_t str_len = strlen(buf);
	
	if (!str_len)
	{
		free(buf);
		return BBSTRING_FROMBUFFER(EmptyArrayMessageBuf, 12);
	}
	
	BBString *str = bbStringFromBytes(buf, str_len);
	free(buf);
	return str;
}

void* bbArrayToCArray(const BBArray *arr)
{
	if (!arr->scales[0] || BBARRAY_ISCOMPLEX(arr)) {
		return NULL;
	}
	
	void *out = bbMemAlloc(arr->size);
	if (!out) {
		return NULL;
	}
	
	memcpy(out, BBARRAY_DATAPTR(arr), arr->size);
	
	return out;
}

BBArray* bbArrayClone(const BBArray *arr)
{
	const BBSize len = arr->scales[0];
	
	if (len < 1) {
		return &bbEmptyArray;
	}
	
	BBArray *out = bb_array_alloc(arr->type, arr->dims, arr->scales);
	
	const BBTypeKey tkey = BBTYPETAG_KEY(arr->type);
	
	if (BBTYPEKEY_ISPRIMITIVE(tkey)) {
		memcpy(BBARRAY_DATAPTR(out), BBARRAY_DATAPTR(arr), len * BBTYPEKEY_SIZE(tkey));
	} else {
				
		BBObject **out_data			= ((BBObjectArray *)out)->data;
		const BBObject **arr_data	= ((BBObjectArray *)arr)->data;
		
		for (arrpos_t pos = 0; pos < len; ++pos)
		{
			out_data[pos] = arr_data[pos];
			BB_RETAIN(out_data[pos]);
		}
	}
	
	return out;
}

BBBool bbArrayEquals(const BBArray *arr, const BBArray *rhs)
{
	if (arr == rhs) {
		return BBTRUE;
	}

	if (arr->scales[0] != rhs->scales[0] ||
		arr->dims != rhs->dims ||
		arr->size != rhs->size ||
		arr->type[0] != rhs->type[0]) {
		return false;
	}
	
	return array_sub_equals(arr, rhs, 0, rhs->scales[0]);
}



BBInt bbArrayFind(const BBArray *arr, const BBArray *sub, BBInt start_pos)
{
	const BBSize arr_len = arr->scales[0];
	const BBSize sub_len = sub->scales[0];
	
	if (sub_len < 1 || sub_len > arr_len) {
		return BBARRAY_NOTFOUND;
	}

	if (start_pos < 1) {
		start_pos = 0;
	} else if (start_pos + sub_len > arr_len) {
		return BBARRAY_NOTFOUND;
	} else if (arr == sub) {
		// When the both array's are the same, the 'sub' one can only
		// be found if the start position for the search is zero
		return !start_pos;
	}	

	for (arrpos_t arr_pos = start_pos; arr_pos < arr_len; ++arr_pos)
	{
		if (array_sub_equals(arr, sub, arr_pos, sub_len)) {
			return arr_pos;
		}
	}
	
	return BBARRAY_NOTFOUND;
}

BBInt bbArrayFindInt(const BBArray *arr, const BBInt num, BBInt start_pos)
{
	const BBTypeKey tkey = BBTYPETAG_KEY(arr->type);	
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 1 || tkey > BBTYPEKEY_INT) {
		return BBARRAY_NOTFOUND;
	}
	
	while (start_pos < 0) {
		start_pos += arr_len;
	}
	if (start_pos >= arr_len) {
		return BBARRAY_NOTFOUND;
	}
	
	if (tkey == BBTYPEKEY_INT)
	{
		const BBInt *data = ((BBIntArray *)arr)->data;	
		
		for (arrpos_t k = start_pos; k < arr_len; ++k)
		{
			if (num == data[k]) {
				return k;
			}
		}
	} else
	if (tkey == BBTYPEKEY_BYTE)
	{
		const BBByte *data = ((BBByteArray *)arr)->data;
		
		for (arrpos_t k = start_pos; k < arr_len; ++k)
		{
			if (num == data[k]) {
				return k;
			}
		}
	} else // Array must be BBShortArray now
	{
		const BBShort *data = ((BBShortArray *)arr)->data;
		
		for (arrpos_t k = start_pos; k < arr_len; ++k)
		{
			if (num == data[k]) {
				return k;
			}
		}
	}
	
	return BBARRAY_NOTFOUND;
}

BBInt bbArrayFindLong(const BBLongArray *arr, const BBLong num, BBInt start_pos)
{
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 1 || BBTYPETAG_KEY(arr->type) != BBTYPEKEY_LONG) {
		return BBARRAY_NOTFOUND;
	}
	
	while (start_pos < 0) {
		start_pos += arr_len;
	}
	if (start_pos >= arr_len) {
		return BBARRAY_NOTFOUND;
	}
	
	const BBLong *data = arr->data;
		
	for (arrpos_t k = start_pos; k < arr_len; ++k)
	{
		if (num == data[k]) {
			return k;
		}
	}
	
	return BBARRAY_NOTFOUND;
}

BBInt bbArrayFindFloat(const BBFloatArray *arr, const BBFloat num, BBInt start_pos)
{
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 1 || BBTYPETAG_KEY(arr->type) != BBTYPEKEY_FLOAT) {
		return BBARRAY_NOTFOUND;
	}
	
	while (start_pos < 0) {
		start_pos += arr_len;
	}
	if (start_pos >= arr_len) {
		return BBARRAY_NOTFOUND;
	}
	
	const BBFloat *data = arr->data;
	
	for (arrpos_t k = 0; k < arr_len; ++k)
	{
		if (num == data[k]) {
			return k;
		}
	}
	
	return BBARRAY_NOTFOUND;
}

BBInt bbArrayFindDouble(const BBDoubleArray *arr, BBDouble num, BBInt start_pos)
{
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 1 || BBTYPETAG_KEY(arr->type) != BBTYPEKEY_DOUBLE) {
		return BBARRAY_NOTFOUND;
	}
	
	while (start_pos < 0) {
		start_pos += arr_len;
	}
	if (start_pos >= arr_len) {
		return BBARRAY_NOTFOUND;
	}
	
	const BBDouble *data = arr->data;
	
	for (arrpos_t k = start_pos; k < arr_len; ++k)
	{
		if (num == data[k]) {
			return k;
		}
	}
	
	return BBARRAY_NOTFOUND;
}

BBInt bbArrayFindObject(const BBObjectArray *arr, const BBObject *obj, BBInt start_pos)
{
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 1 || BBTYPETAG_KEY(arr->type) != BBTYPEKEY_OBJECT) {
		return BBARRAY_NOTFOUND;
	}
	
	while (start_pos < 0) {
		start_pos += arr_len;
	}
	if (start_pos >= arr_len) {
		return BBARRAY_NOTFOUND;
	}
	
	const BBClassCompareFunc cmp_func = obj->clas->Compare;
	
	const BBObject **data = arr->data;
	
	for (arrpos_t k = start_pos; k < arr_len; ++k)
	{
		if (!cmp_func(obj, data[k])) {
			return k;
		}
	}
	
	return BBARRAY_NOTFOUND;
}

BBInt bbArrayFindString(const BBStringArray *arr, const BBString * const str, BBInt start_pos)
{
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 1 || BBTYPETAG_KEY(arr->type) != BBTYPEKEY_STRING) {
		return BBARRAY_NOTFOUND;
	}
	
	while (start_pos < 0) {
		start_pos += arr_len;
	}
	if (start_pos >= arr_len) {
		return BBARRAY_NOTFOUND;
	}
	
	const BBString **data = arr->data;
		
	for (arrpos_t k = start_pos; k < arr_len; ++k)
	{
		if (bbStringEquals(str, data[k])) {
			return k;
		}
	}
	
	return BBARRAY_NOTFOUND;
}

BBBool bbArrayStartsWith(const BBArray *arr, const BBArray *rhs)
{
	const BBSize rhs_len = rhs->scales[0];
	
	if (arr->scales[0] < rhs_len || arr->type[0] != rhs->type[0]) {
		return BBFALSE;
	} else if (arr == rhs) {
		return BBTRUE;
	}

	const BBTypeKey arr_tkey = BBTYPETAG_KEY(arr->type);
	
	if (!BBTYPEKEY_ISCOMPLEX(arr_tkey)) {
		
		return !memcmp(BBARRAY_DATAPTR(arr),
					  BBARRAY_DATAPTR(rhs),
					  rhs_len  * BBTYPEKEY_SIZE(arr_tkey)
					 );
	}
	
	if (arr_tkey == BBTYPEKEY_STRING) {
		
		const BBString **arr_data = ((BBStringArray *)arr)->data;
		const BBString **rhs_data = ((BBStringArray *)rhs)->data;
		
		for (arrpos_t k = 0; k < rhs_len; ++k)
		{
			if (!bbStringEquals(arr_data[k], rhs_data[k])) {
				return BBFALSE;
			}
		}
		
	} else if (arr_tkey == BBTYPEKEY_OBJECT) {
		
		const BBObject **arr_data = ((BBObjectArray *)arr)->data;
		const BBObject **rhs_data = ((BBObjectArray *)rhs)->data;
		
		BBClassCompareFunc cmp_func;
		
		for (arrpos_t k = 0; k < rhs_len; ++k)
		{
			cmp_func = arr_data[k]->clas->Compare;
			
			if (cmp_func(arr_data[k], rhs_data[k])) {
				return BBFALSE;
			}
		}
		
	} else if (arr_tkey == BBTYPEKEY_ARRAY) {
		
		const BBArray **arr_data = ((BBArrayArray *)arr)->data;
		const BBArray **rhs_data = ((BBArrayArray *)rhs)->data;
		
		for (arrpos_t k = 0; k < rhs_len; ++k)
		{
			if (!bbArrayEquals(arr_data[k], rhs_data[k])) {
				return BBFALSE;
			}
		}
	}
	
	return BBTRUE;
}

BBBool bbArrayEndsWith(const BBArray *arr, const BBArray *rhs)
{
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < rhs->scales[0] || arr->type[0] != rhs->type[0]) {
		return BBFALSE;
	} else if (arr == rhs) {
		return BBTRUE;
	}
	
	const BBTypeKey arr_tkey = BBTYPETAG_KEY(arr->type);
	
	if (!BBTYPEKEY_ISCOMPLEX(arr_tkey)) {
		
		size_t cmp_len = BBTYPEKEY_SIZE(arr_tkey) * (arr_len - rhs->scales[0]);
		
		return !memcmp(
					((char *)BBARRAY_DATAPTR(arr)) + cmp_len,
					BBARRAY_DATAPTR(rhs),
					rhs->scales[0] * BBTYPEKEY_SIZE(arr_tkey)
				);
	}
	
	if (arr_tkey == BBTYPEKEY_STRING) {
		
		const BBString **arr_data = ((BBStringArray *)arr)->data;
		const BBString **rhs_data = ((BBStringArray *)rhs)->data;
		
		for (arrpos_t k = rhs->scales[0]; k > arr_len; ++k)
		{
			if (!bbStringEquals(arr_data[k], rhs_data[k])) {
				return BBFALSE;
			}
		}
		
	} else if (arr_tkey == BBTYPEKEY_OBJECT) {
		
		const BBObject **arr_data = ((BBObjectArray *)arr)->data;
		const BBObject **rhs_data = ((BBObjectArray *)rhs)->data;
		
		const BBClassCompareFunc cmp_func = arr->clas->Compare;
		
		for (arrpos_t k = rhs->scales[0]; k > arr_len; ++k)
		{
			if (cmp_func(arr_data[k], rhs_data[k])) {
				return BBFALSE;
			}
		}
		
	} else if (arr_tkey == BBTYPEKEY_ARRAY) {
		
		const BBArray **arr_data = ((BBArrayArray *)arr)->data;
		const BBArray **rhs_data = ((BBArrayArray *)rhs)->data;
		
		for (arrpos_t k = rhs->scales[0]; k > arr_len; ++k)
		{
			if (!bbArrayEquals(arr_data[k], rhs_data[k])) {
				return BBFALSE;
			}
		}
	}
	
	return BBTRUE;
}



BBBool bbArrayContains(const BBArray *arr, const BBArray *sub)
{
	return bbArrayFind(arr, sub, 0) != BBARRAY_NOTFOUND;
}

BBBool bbArrayContainsInt(const BBArray *arr, BBInt num)
{
	return bbArrayFindInt(arr, num, 0) != BBARRAY_NOTFOUND;
}

BBBool bbArrayContainsLong(const BBLongArray *arr, BBLong num)
{
	return bbArrayFindLong(arr, num, 0) != BBARRAY_NOTFOUND;
}

BBBool bbArrayContainsFloat(const BBFloatArray *arr, BBFloat num)
{
	return bbArrayFindFloat(arr, num, 0) != BBARRAY_NOTFOUND;
}

BBBool bbArrayContainsDouble(const BBDoubleArray *arr, BBDouble num)
{
	return bbArrayFindDouble(arr, num, 0) != BBARRAY_NOTFOUND;
}

BBBool bbArrayContainsObject(const BBObjectArray *arr, const BBObject *obj)
{
	return bbArrayFindObject(arr, obj, 0) != BBARRAY_NOTFOUND;
}

BBBool bbArrayContainsString(const BBStringArray *arr, const BBString *str)
{
	return bbArrayFindString(arr, str, 0) != BBARRAY_NOTFOUND;
}


//	REVERSE //

BBByteArray* bb_bytearray_reverse(const BBByteArray *arr)
{
	BB_ASSERT(arr->type[0] == 'b', "Invalid array type. (Expected BBByte)");
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 2) {
		return arr;
	}
	
	BBByteArray *out		= bb_array_alloc("b", arr->dims, arr->scales);
	
	BBByte *out_data		= out->data;
	BBByte *arr_data		= arr->data + (arr_len - 1);
	
	for (arrpos_t k = 0; k < arr_len; ++k)
	{
		out_data[k] = arr_data[-k];
	}
	
	return out;
}

BBShortArray* bb_shortarray_reverse(const BBShortArray *arr)
{
	BB_ASSERT(arr->type[0] == 's', "Invalid array type. (Expected BBShort)");
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 2) {
		return arr;
	}
	
	BBShortArray *out		= bb_array_alloc("s", arr->dims, arr->scales);
	
	BBShort *out_data		= out->data;
	BBShort *arr_data		= arr->data + (arr_len - 1);
	
	for (arrpos_t k = 0; k < arr_len; ++k)
	{
		out_data[k] = arr_data[-k];
	}
	
	return out;
}

BBIntArray* bb_intarray_reverse(const BBIntArray *arr)
{
	BB_ASSERT(arr->type[0] == 'i', "Invalid array type. (Expected BBInt)");
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 2) {
		return arr;
	}
	
	BBIntArray *out		= bb_array_alloc("i", arr->dims, arr->scales);
	
	BBInt *out_data		= out->data;
	BBInt *arr_data		= arr->data + (arr_len - 1);
	
	for (arrpos_t k = 0; k < arr_len; ++k)
	{
		out_data[k] = arr_data[-k];
	}
	
	return out;
}

BBLongArray* bb_longarray_reverse(const BBLongArray *arr)
{
	BB_ASSERT(arr->type[0] == 'l', "Invalid array type. (Expected BBInt)");
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 2) {
		return arr;
	}
	
	BBLongArray *out		= bb_array_alloc("l", arr->dims, arr->scales);
	
	BBLong *out_data		= out->data;
	BBLong *arr_data		= arr->data + (arr_len - 1);
	
	for (arrpos_t k = 0; k < arr_len; ++k)
	{
		out_data[k] = arr_data[-k];
	}
	
	return out;
}

BBFloatArray* bb_floatarray_reverse(const BBFloatArray *arr)
{
	BB_ASSERT(arr->type[0] == 'f', "Invalid array type. (Expected BBFloat)");
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 2) {
		return arr;
	}
	
	BBFloatArray *out	= bb_array_alloc("f", arr->dims, arr->scales);
	
	BBFloat *out_data	= out->data;
	BBFloat *arr_data	= arr->data + (arr_len - 1);
	
	for (arrpos_t k = 0; k < arr_len; ++k)
	{
		out_data[k] = arr_data[-k];
	}
	
	return out;
}

BBDoubleArray* bb_doublearray_reverse(const BBDoubleArray *arr)
{
	BB_ASSERT(arr->type[0] == 'd', "Invalid array type. (Expected BBDouble)");
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 2) {
		return arr;
	}
	
	BBDoubleArray *out	= bb_array_alloc("d", arr->dims, arr->scales);
	
	BBDouble *out_data	= out->data;
	BBDouble *arr_data	= arr->data + (arr_len - 1);
	
	for (arrpos_t k = 0; k < arr_len; ++k)
	{
		out_data[k] = arr_data[-k];
	}
	
	return out;
}

// todo: Maybe calculate element size and copy per byte?
BBArray* bb_pointerarray_reverse(const BBArray *arr)
{
	BB_ASSERT(arr->type[0] == '*', "Invalid array type. (Expected pointer)");
	
	BB_ASSERT(
		sizeof (int *) == sizeof (BBByte *) &&
		sizeof (int *) == sizeof (BBShort *) &&
		sizeof (int *) == sizeof (BBInt *) &&
		sizeof (int *) == sizeof (BBLong *) &&
		sizeof (int *) == sizeof (BBFloat *) &&
		sizeof (int *) == sizeof (BBDouble *) &&
		sizeof (int *) == sizeof (BBByte **),
		"Error: Some pointers are incompatible to int*."
	);

	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 2) {
		return arr;
	}
	
	BBArray *out			= bb_array_alloc(arr->type, arr->dims, arr->scales);
	
	const BBSize el_size = arr->size / arr_len;
	
	char *out_data		= BBARRAY_DATAPTR(out);
	char *arr_data		= BBARRAY_DATAPTR(arr) + ((arr_len-1) * el_size);
	
	for (arrpos_t k = 0; k < arr_len; ++k)
	{
		memcpy(out_data + k, arr_data - k, el_size);
	}
	
	return out;
}

BBObjectArray* bb_objectarray_reverse(BBObjectArray *arr)
{
	BB_ASSERT(BBARRAY_ISCOMPLEX(arr), "Invalid array type. (Expected BBObject*, BBString* or BBArray*)");
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 2) {
		return arr;
	}
	
	BBObjectArray *out	= bb_array_alloc(arr->type, arr->dims, arr->scales);
	
	BBObject **out_data	= out->data;
	BBObject **arr_data	= arr->data + (arr_len - 1);
	
	for (arrpos_t k = 0; k < arr_len; ++k)
	{
		out_data[k] = arr_data[-k];
		BB_INCREFS(out_data[k]);
	}
	
	return out;
}

BBArray* bbArrayReverse(const BBArray *arr)
{
	if (BBARRAY_ISNUMERIC(arr)) {
		switch (arr->type[0]) {
		case 'i':
			return bb_intarray_reverse(arr);
		case 'f':
			return bb_floatarray_reverse(arr);
		case 'd':
			return bb_doublearray_reverse(arr);
		case 'b':
			return bb_bytearray_reverse(arr);
		case 's':
			return bb_shortarray_reverse(arr);
		case 'l':
			return bb_longarray_reverse(arr);
		}
	}
	else if (BBARRAY_ISCOMPLEX(arr)) {
		return bb_objectarray_reverse(arr);		
	}
	else if (arr->type[0] == '*') {
		return bb_pointerarray_reverse(arr);		
	}
	
	return BBNULLARRAY;
}


BBArray* bbArrayRotate(const BBArray *arr, BBInt count)
{
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 2 || count == 0 || count == arr_len) {
		return arr;
	}
	
	count = -(count % arr_len);
	
	if (count < 0) {
		count += arr_len;
	}
	
	const BBTypeKey typekey		= BBTYPETAG_KEY(arr->type);
	const BBSize element_size	= BBTYPEKEY_SIZE(typekey);
	
	char *arr_data	= BBARRAY_DATAPTR(arr);
	
	BBArray *out	= bb_array_alloc(arr->type, arr->dims, arr->scales);
	char *out_data	= BBARRAY_DATAPTR(out);
	
	for (arrpos_t pos = 0; pos < arr_len; ++pos, ++count)
	{
		if (count >= arr_len) {
			count = 0;
		}
		
		memcpy(out_data + (pos * element_size),
			   arr_data + ((count) * element_size),
			   element_size);
	}
	
	return out;
}

//	FILTER //

BBIntArray* bb_intarray_filter(arr, filter)
 const BBIntArray *arr;
 BBBool (*filter)(BBInt);
{
	const BBSize arr_len = arr->scales[0];
	
	if (!arr_len) {
		return &bbEmptyArray;
	}
	
	BBSize hits[arr_len];
	BBSize num_hits = 0;
	
	arrpos_t pos;
	
	for (pos = 0; pos < arr_len; ++pos)
	{
		// Register all valid fields
		if ((*filter)(arr->data[pos])) {
			hits[num_hits++] = pos;
		}
	}
	
	if (!num_hits) {
		return &bbEmptyArray;
	}
	
	BBIntArray *out = bb_array_alloc1d("i", num_hits);

	for (pos = 0; pos < num_hits; ++pos)
	{
		out->data[pos] = arr->data[hits[pos]];
	}
	
	return out;
}

BBLongArray* bb_longarray_filter(arr, filter)
 const BBLongArray *arr;
 BBBool (*filter)(BBLong);
{
	const BBSize arr_len = arr->scales[0];
	
	if (!arr_len) {
		return &bbEmptyArray;
	}
	
	BBSize hits[arr_len];
	BBSize num_hits = 0;
	
	arrpos_t pos;
	
	for (pos = 0; pos < arr_len; ++pos)
	{
		// Register all valid fields
		if ((*filter)(arr->data[pos])) {
			hits[num_hits++] = pos;
		}
	}
	
	if (!num_hits) {
		return &bbEmptyArray;
	}
	
	BBLongArray *out = bb_array_alloc1d("l", num_hits);

	for (pos = 0; pos < num_hits; ++pos)
	{
		out->data[pos] = arr->data[hits[pos]];
	}
	
	return out;
}

BBFloatArray* bb_floatarray_filter(arr, filter)
 const BBFloatArray *arr;
 BBBool (*filter)(BBFloat);
{
	const BBSize arr_len = arr->scales[0];
	
	if (!arr_len) {
		return &bbEmptyArray;
	}
	
	BBSize hits[arr_len];
	BBSize num_hits = 0;
	
	arrpos_t pos;
	
	for (pos = 0; pos < arr_len; ++pos)
	{
		// Register all valid fields
		if ((*filter)(arr->data[pos])) {
			hits[num_hits++] = pos;
		}
	}
	
	if (!num_hits) {
		return &bbEmptyArray;
	}
	
	BBFloatArray *out = bb_array_alloc1d("f", num_hits);

	for (pos = 0; pos < num_hits; ++pos)
	{
		out->data[pos] = arr->data[hits[pos]];
	}
	
	return out;
}

BBDoubleArray* bb_doublearray_filter(arr, filter)
 const BBDoubleArray *arr;
 BBBool (*filter)(BBDouble);
{
	const BBSize arr_len = arr->scales[0];
	
	if (!arr_len) {
		return &bbEmptyArray;
	}
	
	BBSize hits[arr_len];
	BBSize num_hits = 0;
	
	arrpos_t pos;
	
	for (pos = 0; pos < arr_len; ++pos)
	{
		// Register all valid fields
		if ((*filter)(arr->data[pos])) {
			hits[num_hits++] = pos;
		}
	}
	
	if (!num_hits) {
		return &bbEmptyArray;
	}
	
	BBFloatArray *out = bb_array_alloc1d("d", num_hits);

	for (pos = 0; pos < num_hits; ++pos)
	{
		out->data[pos] = arr->data[hits[pos]];
	}
	
	return out;
}

BBObjectArray* bb_objectarray_filter(arr, filter)
 const BBObjectArray *arr;
 BBBool (*filter)(BBObject *);
{
	const BBSize arr_len = arr->scales[0];
	
	if (!arr_len || !BBARRAY_ISCOMPLEX(arr)) {
		return &bbEmptyArray;
	}
	
	BBSize hits[arr_len];
	BBSize num_hits = 0;
	
	arrpos_t pos;
	
	for (pos = 0; pos < arr_len; ++pos)
	{
		// Register all valid fields
		if ((*filter)(arr->data[pos])) {
			hits[num_hits++] = pos;
		}
	}
	
	if (!num_hits) {
		return &bbEmptyArray;
	}
	
	BBObjectArray *out = bb_array_alloc1d(arr->type, num_hits);

	for (pos = 0; pos < num_hits; ++pos) {
		out->data[pos] = arr->data[hits[pos]];
		BB_INCREFS(out->data[pos]);
	}
	
	return out;
}

BBArray* bbArrayFilter(const BBArray *arr, BBBool (*filter)(void *))
{
	switch (arr->type[0]) {
	case 'i':
		return bb_intarray_filter(arr, (BBBool (*)(BBInt))filter);
	case 'f':
		return bb_floatarray_filter(arr, (BBBool (*)(BBFloat))filter);
	case 'd': 
		return bb_doublearray_filter(arr, (BBBool (*)(BBDouble))filter);
	case 'l':
		return bb_longarray_filter(arr, (BBBool (*)(BBLong))filter);
	case ':': case '$': case '[':
		return bb_objectarray_filter(arr, (BBBool (*)(BBObject *))filter);
	}
	
	return BBNULLARRAY;
}


//	MAP //

BBIntArray* bb_intarray_map(arr, callback)
 const BBIntArray* arr;
 BBInt (*callback)(BBInt);
{
	const BBSize arr_len = arr->scales[0];
	
	if (!arr_len) {
		return BBNULLARRAY;
	}
	
	BBIntArray* out = bb_array_alloc("i", arr->dims, arr->scales);
	
	for (arrpos_t pos = 0; pos < arr_len; ++pos) {
		out->data[pos] = (*callback)(arr->data[pos]);
	}
	
	return out;
}

BBLongArray* bb_longarray_map(arr, callback)
 const BBLongArray* arr;
 BBLong (*callback)(BBLong);
{
	const BBSize arr_len = arr->scales[0];
	
	if (!arr_len) {
		return BBNULLARRAY;
	}
	
	BBIntArray* out = bb_array_alloc("l", arr->dims, arr->scales);
	
	for (arrpos_t pos = 0; pos < arr_len; ++pos) {
		out->data[pos] = (*callback)(arr->data[pos]);
	}
	
	return out;
}

BBFloatArray* bb_floatarray_map(arr, callback)
 const BBFloatArray* arr;
 BBFloat (*callback)(BBFloat);
{
	const BBSize arr_len = arr->scales[0];
	
	if (!arr_len) {
		return BBNULLARRAY;
	}
	
	BBFloatArray* out = bb_array_alloc("f", arr->dims, arr->scales);
	
	for (arrpos_t pos = 0; pos < arr_len; ++pos) {
		out->data[pos] = (*callback)(arr->data[pos]);
	}
	
	return out;
}

BBDoubleArray* bb_doublearray_map(arr, callback)
 const BBDoubleArray* arr;
 BBDouble (*callback)(BBDouble);
{
	const BBSize arr_len = arr->scales[0];
	
	if (!arr_len) {
		return BBNULLARRAY;
	}
	
	BBDoubleArray* out = bb_array_alloc("d", arr->dims, arr->scales);
	
	for (arrpos_t pos = 0; pos < arr_len; ++pos) {
		out->data[pos] = (*callback)(arr->data[pos]);
	}
	
	return out;
}

BBObjectArray* bb_objectarray_map(arr, callback)
 BBObjectArray* arr;
 BBObject* (*callback)(BBObject *);
{
	const BBSize arr_len = arr->scales[0];
	
	if (!arr_len || !BBARRAY_ISCOMPLEX(arr)) {
		return BBNULLARRAY;
	}
	
	BBObjectArray* out = bb_array_alloc(arr->type, arr->dims, arr->scales);
	
	for (arrpos_t pos = 0; pos < arr_len; ++pos) {
		out->data[pos] = (*callback)(arr->data[pos]);
	}
	
	return out;
}

BBArray* bbArrayMap(BBArray *arr, void* (*callback)(void *))
{
	switch (arr->type[0]) {
	case 'i':
		return bb_intarray_map(arr, (BBInt (*)(BBInt))callback);
	case 'f':
		return bb_floatarray_map(arr, (BBFloat (*)(BBFloat))callback);
	case 'd': 
		return bb_doublearray_map(arr, (BBDouble (*)(BBDouble))callback);
	case 'l':
		return bb_longarray_map(arr, (BBLong (*)(BBLong))callback);
	case ':': case '$': case '[':
		return bb_objectarray_map(arr, (BBObject* (*)(BBObject *))callback);
	}
	
	return BBNULLARRAY;
}
