
#include "blitz_types.h"

#include "blitz_object.h"
#include "blitz_string.h"
#include "blitz_array.h"
#include "exceptions/blitz_nullfunctionexception.h"

const char *bbVoidTypeTag	= "?";
const char *bbByteTypeTag	= "b";
const char *bbShortTypeTag	= "s";
const char *bbIntTypeTag		= "i";
const char *bbFloatTypeTag	= "f";
const char *bbDoubleTypeTag	= "d";

const char *bbBytePtrTypeTag	= "*b";

const char *bbObjectTypeTag	= ":Object";
const char *bbStringTypeTag	= "$";

BBBool bb_typetag_isvalid(const char *type)
{
#	define NUM_TAGS			12
	// All valid TypeTags ordered by frequency of usage.
	static const char all_tags[] = {
		'i', 'f', 'd', ':', '$', '[',
		'b', 's', 'l', '*', '(', '?'
	};
	
	for (int t = 0; t < NUM_TAGS; ++t)
	{
		if (all_tags[t] == type[0]) {
			return BBTRUE;
		}
	}
	
	return BBFALSE;
}

char* bb_typetag_identifier(const char *type)
{
	if (!bb_typetag_isvalid(type)) {
		return NULL;
	}
	
	const BBTypeKey tkey = bbToTypeKeyData[type[0]];
	
	char *ident			= NULL;
	size_t ident_len		= 0;
	
	// All these TypeKeys have constant names.
	if (BBTYPEKEY_ISNUMERIC(tkey) || tkey == BBTYPEKEY_STRING ||
		!strcmp(type, bbObjectTypeTag)) {

		ident_len	= strlen(bbTypeKeyInfo[tkey].ident);
		ident		= malloc(ident_len + 1);
		
		strncpy(ident, bbTypeKeyInfo[tkey].ident, ident_len);
		ident[ident_len] = '\0';
		
		return ident;
	}
	
	if (tkey == BBTYPEKEY_OBJECT) {
		
		ident_len	= strlen(type) - 1;
		ident		= malloc(ident_len + 1);
		
		strncpy(ident, type + 1, ident_len);
		ident[ident_len] = '\0';
		
		return ident;
	}
	
	if (tkey == BBTYPEKEY_POINTER) {
		
		// Only pointers to primitive data types are supported
		// which have a TypeTag of length 1. Therefore the TypeTag's
		// length minus one equals the level of pointers.
		const BBSize num_ptrs	= strlen(type) - 1;
		
		// Get the TypeKey of the pointed to value.
		const BBTypeKey val_tkey	= bbToTypeKeyData[ type[num_ptrs] ];
		
		ident_len	= strlen(BBTYPEKEY_IDENT(val_tkey))
					+ 1					/* For space between identifier and "Ptr" */
					+ (num_ptrs * 3)		/* For each "Ptr" keyword */
					+ (num_ptrs - 1);	/* For space between each "Ptr" */				
		
		ident		= malloc(ident_len + 1);
		strcat(ident, BBTYPEKEY_IDENT(val_tkey));
		
		for (int ptr_id = 0; ptr_id < num_ptrs; ++ptr_id)
		{
			strcat(ident, " Ptr");
		}
		
		return ident;
	}
	
	if (tkey == BBTYPEKEY_FUNCTION) {
	
		size_t len = strlen(type);
		
		if (len <= 2) {
			ident	= malloc(11);
			strcpy(ident, "Function()");
			return ident;
		}
		
		// At least "()" can be ignored as parameters/return.
		char *args[len - 2];
		char *ret_ident;
		
		size_t last_pos	= (strrchr(type, ')') + 1) - type;
		
		/*
		// Check if the function has a return type. [ (x,y,z)$TYPE ]
		if (last_pos < len) {
			ret_ident = bb_typetag_identifier(type + last_pos);
		} else {
			ret_ident = NULL;
		}*/
		
		size_t pos;
		
		// Check if the function has parameters. [ ()$TYPE == no params ]
		if (last_pos > 2) {
		
			char *tag = strtok(type + 1, ",");
			
			while (tag != NULL)
			{
				if (tag - type == last_pos) {
					ret_ident			= bb_typetag_identifier(tag);
				} else {
					args[ident_len++]	= bb_typetag_identifier(tag);					
				}
				
				tag = strtok(NULL, ",)");
			}

/*
			char tag[len];
			last_pos = 1;
		
			// Add all parameters. Start at "1" because at "0" it's "(" anyway.
			for (pos = 1; pos < len; ++pos)
			{
				if (type[pos] == ',' || type[pos] == ')') {
					
					strncpy(tag, type + last_pos, pos - last_pos);
					tag[pos - last_pos] = '\0';
				
					args[ident_len] = bb_typetag_identifier(tag);

					printf("last_pos = %i pos = %i\n", last_pos, pos);					
					last_pos = pos;
					++ident_len;
				}
			}
*/
			
			len			= ident_len;
			ident_len	= 0;
			
			// Reserve memory for all parameters.
			for (pos = 0; pos < len; ++pos)
			{
				ident_len += strlen(args[pos]);
			}
			
		}	// end: Function has params
		
		ident_len		+= 8;	// For "Function"
		
		// Reserve memory for return type, if declated.
		if (ret_ident) {
			ident_len	+= 1	// For ":" after function
						 + strlen(ret_ident);
		}
		
		// Reserve 
		if (len) {
			ident_len	+= len;	// For space between every parameter
		}
		
		ident		= malloc(ident_len + 1);
		
		strcpy(ident, "Function");
		
		// Add return type, if declared. 
		if (ret_ident) {
			strcat(ident, ":");
			strcat(ident, ret_ident);
		}
		
		strcat(ident, "(");
		
		// Add parameters, if declared.
		if (len >= 1) {
		
			strcat(ident, args[0]);
			
			for (pos = 1; pos < len; ++pos)
			{
				strcat(ident, ", ");	
				strcat(ident, args[pos]);
			}
		
		}
		
		strcat(ident, ")");
		
		return ident;
	}
}

// Maps the first character of each valid type tag to a type key.
const BBTypeKey bbToTypeKeyData[] =
{
	['?'] = BBTYPEKEY_VOID,
	['b'] = BBTYPEKEY_BYTE,
	['s'] = BBTYPEKEY_SHORT,
	['i'] = BBTYPEKEY_INT,
	['l'] = BBTYPEKEY_LONG,
	['f'] = BBTYPEKEY_FLOAT,
	['d'] = BBTYPEKEY_DOUBLE,
	
	['*'] = BBTYPEKEY_POINTER,
	['('] = BBTYPEKEY_FUNCTION,
	
	['$'] = BBTYPEKEY_STRING,
	[':'] = BBTYPEKEY_OBJECT,
	['['] = BBTYPEKEY_ARRAY
};

const char *bbToTypeTagData[] =
{
	[BBTYPEKEY_BYTE]		= "b",
	[BBTYPEKEY_SHORT]	= "s",
	[BBTYPEKEY_INT]		= "i",
	[BBTYPEKEY_LONG]		= "l",
	[BBTYPEKEY_FLOAT]	= "f",
	[BBTYPEKEY_DOUBLE]	= "d",
	
	[BBTYPEKEY_POINTER]	= "*",
	[BBTYPEKEY_FUNCTION]	= "(",
	
	[BBTYPEKEY_STRING]	= "$",
	[BBTYPEKEY_OBJECT]	= ":",
	[BBTYPEKEY_ARRAY]	= "["
};

const BBTypeKeyInfo bbTypeKeyInfo[] =
{
	[BBTYPEKEY_BYTE] = {
		.size	= sizeof (BBByte),
		.init	= (void *)0,
		.ident	= "Byte"
	},
	
	[BBTYPEKEY_SHORT] = {
		.size	= sizeof (BBShort),
		.init	= (void *)0,
		.ident	= "Short"
	},
	
	[BBTYPEKEY_INT] = {
		.size	= sizeof (BBInt),
		.init	= (void *)0,
		.ident	= "Int"
	},
	
	[BBTYPEKEY_LONG] = {
		.size	= sizeof (BBLong),
		.init	= (void *)0,
		.ident	= "Long"
	},
	
	[BBTYPEKEY_FLOAT] = {
		.size	= sizeof (BBFloat),
		.init	= (void *)0,
		.ident	= "Float"
	},
	
	[BBTYPEKEY_DOUBLE] = {
		.size	= sizeof (BBDouble),
		.init	= (void *)0,
		.ident	= "Double"
	},
	
	[BBTYPEKEY_POINTER]	= {
		.size	= sizeof (int*),
		.init	= NULL,
		.ident	= "Ptr"
	},
	
	[BBTYPEKEY_FUNCTION] = {
		.size	= sizeof (void* (*)),
		.init	= (void *)&brl_blitz_NullFunctionError,
		.ident	= "Function"
	},
	
	//	Complex
	
	[BBTYPEKEY_OBJECT] = {
		.size	= sizeof (BBObject*),
		.init	= (void *)&bbNullObject,
		.ident	= "Object"
	},
	
	[BBTYPEKEY_STRING] = {
		.size	= sizeof (BBString*),
		.init	= (void *)&bbEmptyString,
		.ident	= "String"
	},
	
	[BBTYPEKEY_ARRAY] = {
		.size	= sizeof (BBArray*),
		.init	= (void *)&bbEmptyArray,
		.ident	= "Array"
	}
};

BBInt bbConvertToInt(const void *value, const char *tag)
{
	switch (tag[0])
	{
	case 'b':
		return *(BBByte *)value;
	case 's':
		return *(BBShort *)value;
	case 'i':
		return *(BBInt *)value;
	case 'f':
		return *(BBFloat *)value;
	case 'd':
		return *(BBDouble *)value;
	case '$':
		return bbStringToInt(*(BBString **)value);
	}
	
	return 0;
}

BBDouble bbConvertToFloat(const void *value, const char *tag)
{
	switch (tag[0])
	{
	case 'b':
		return *(BBByte *)value;
	case 's':
		return *(BBShort *)value;
	case 'i':
		return *(BBInt *)value;
	case 'f':
		return *(BBFloat *)value;
	case 'd':
		return *(BBDouble *)value;
	case '$':
		return bbStringToFloat(*(BBString **)value);
	}
	
	return 0.0;
}

BBString* bbConvertToString(void *value, const char *tag)
{
	switch (tag[0])
	{
	case 'b':
		return bbStringFromInt(*(BBByte *)value);
	case 's':
		return bbStringFromInt(*(BBShort *)value);
	case 'i':
		return bbStringFromInt(*(BBInt *)value);
	case 'f':
		return bbStringFromFloat(*(BBFloat *)value);
	case 'd':
		return bbStringFromFloat(*(BBDouble *)value);
	case '$':
		return *(BBString **)value;
	}
	
	return 0;
}
